/* Test various hash functions, 2005-03-09 */
/* Public domain, by C.B. Falconer */
/* Attribution appreciated */
/* Comments and bugs to cbfalconer@worldnet.att.net */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "ggets.h"
#include "hashlib.h"
#include "libhash.h"

/* diddle this to suit your system */
#define REPS 6  /* to get suitable resolution on timing */
#define MAXWORD 80

/* This will be the head of a list holding the input file */
static struct slist {
   struct slist *next;
   char         *datum;
   unsigned long linenum;
   unsigned long duplicates;
} *filedata;

static unsigned long  lines; /* input line counter */

/* The string hashing functions under test */
typedef unsigned long shashfn(const char *);

/* ------------------- */

/* This is very likely to be non-portable */
/* DOES NOT check fp open for reading     */
/* NULL fp is considered a keyboard here! */
static int akeyboard(FILE *fp)
{
#ifndef __TURBOC__  /* Turbo C is peculiar */
# ifdef __STDC__
  /* This dirty operation allows gcc -ansi -pedantic */
  extern int fileno(FILE *fp);
  extern int isatty(int fn);
# endif
#endif
   return ((fp != NULL) && isatty(fileno(fp)));
} /* akeyboard */

/* ------------------- */

static void help(void)
{
   puts("No input redirection detected");
   puts("Usage: hshftst <testfile");
} /* help */

/* ------------------- */

static double timestrinput(FILE *f)

{
   clock_t       cstart, cend;
   struct slist *node;
   char         *ln;

   cstart = clock();
   while (0 == fggets(&ln, f)) {
      if (!(node = malloc(sizeof(*node)))) break;
      node->datum = ln;
      node->next = filedata;
      filedata = node;
      lines++;
      node->linenum = lines;
   }
   cend = clock();
   return (double)(cend - cstart) / CLOCKS_PER_SEC;
} /* timestrinput */

/* ------------------- */

static unsigned long nullfn(const char *s)
{
   return 0UL;
} /* nullfn */

/* ------------------- */

static double timehashing(struct slist *root, shashfn h)
{
   clock_t       cstart, cend;
   struct slist *node;
   unsigned long hash;
   int           i;

   cstart = clock();
   for (i = 0; i < REPS; i++) {
      node = root;
      while (node) {
         hash = h(node->datum);
         node = node->next;
      }
   }
   cend = clock();
   return (double)(cend - cstart) / CLOCKS_PER_SEC / REPS;
} /* timehashing */

/* ========= hashlib auxiliaryies ========== */

/* Auxiliary functions needed to operate the hash table */
/* See hashlib.h for detailed specifications */

/* Here we are simply going to store the struct slist that
   was created earlier, and give up our control of its storage.
   We already have a handle on the whole thing with the global
   filedata pointer.  This avoid the effort of a deep copy */
static void *hdupex(void *item)
{
   struct slist *node = item;

   node->duplicates = 0; /* to count insertions */
   return item;
} /* hdupex */

/* ------------------- */

/* Similarly we want to keep the stored input data available
   for the next test, and not run into the O(n*n) behaviour of
   some systems free function, so we will just make this
   reversal of hdupe a null operation */
static void hfreex(void *item)
{
   return;
} /* hfreex */

/* ------------------- */

/* here all we need do is dig down to the string pointers */
static int hcmp(void *left, void *right)
{
   struct slist *litem = left,
                *ritem = right;

   return strcmp(litem->datum, ritem->datum);
} /* hcmp */

/* The various actual hashing functions are in the libhash
   library, and are the items to be tested.  They need some
   slight indirection */

/* these customize the following two functions and are set
   by hashopen function */

static shashfn *hasher, *rehasher;

/* ------------------- */

static unsigned long hhash(void *item)
{
   struct slist *node = item;

   return hasher(node->datum);
} /* hhash */

/* ------------------- */

static unsigned long hrehash(void *item)
{
   struct slist *node = item;

   return rehasher(node->datum);
} /* hrehash */

/* ========= End of hashlib auxiliaryies ========== */

/* ------------------- */

/* Open and create a hashtable that uses the passed hash
   functions.  The other facets of the table will not change */
static hshtbl *hashopen(shashfn primary, shashfn secondary,
                        hshdupfn hdupex, hshfreefn hfreex)
{
   hasher = primary; rehasher = secondary;
   return hshinit(hhash, hrehash,
                  hcmp,
                  hdupex, hfreex,
                  0);
} /* hashopen */

/* ------------------- */

static void showstats(hshtbl *h)
{
   hshstats stats;

   stats = hshstatus(h);
   printf("  Entries = %lu Probes = %lu Misses = %lu\n",
           stats.hentries, stats.probes, stats.misses);
} /* showstats */

/* ------------------- */

static double evaluatehashing(hshtbl *h, struct slist *root)
{
   clock_t       cstart, cend;
   struct slist *node, *inserted;
   unsigned long duplines;

   duplines = 0;
   cstart = clock();
   node = root;
   while (node) {
      inserted = hshinsert(h, node);
      /* The following is peculiar to this test program */
      if (inserted->duplicates) duplines++;
      else if (inserted != node) {
         puts("Something is wrong");
         break;
      }
      inserted->duplicates++;
      node = node->next;
   }
   cend = clock();
   printf("  %lu duplicated lines\n", duplines);
   showstats(h);
   return (double)(cend - cstart) / CLOCKS_PER_SEC;
} /* evaluatehashing */

/* ------------------- */

static double timeparsewords(struct slist *root)
{
   clock_t       cstart, cend;
   struct slist *node;
   char          buffer[MAXWORD];
   char         *ptr;
   int           ix;
   unsigned int  wdcount;

   cstart = clock();
   node = root;
   wdcount = 0;
   while (node) {
      ptr = node->datum;
      while (1 == sscanf(ptr, "%s%n", buffer, &ix)) {
         ptr += ix; wdcount++;
      }
      node = node->next;
   }
   cend = clock();
   printf("Found %u words\n", wdcount);
   return (double)(cend - cstart) / CLOCKS_PER_SEC;
} /* timeparsewords */

/* ------------------- */

static double timeparsehashwds(struct slist *root, shashfn h)
{
   clock_t       cstart, cend;
   struct slist *node;
   char          buffer[MAXWORD];
   char         *ptr;
   int           ix;

   cstart = clock();
   node = root;
   while (node) {
      ptr = node->datum;
      while (1 == sscanf(ptr, "%s%n", buffer, &ix)) {
         ptr += ix;
         h(buffer);
      }
      node = node->next;
   }
   cend = clock();
   return (double)(cend - cstart) / CLOCKS_PER_SEC;
} /* timeparsehashwds */

/* ========= More  hashlib auxiliaryies ========== */

/* Auxiliary functions needed to operate the hash table */
/* See hashlib.h for detailed specifications */

/* Unlike the earlier version, to store words we need to make
   the actual deep copies.  The free will also have to be real
*/
static void *hdupewd(void *item)
{
   struct slist *node = item;
   struct slist *newnode;

   if ((newnode = malloc(sizeof *newnode))) {
      if (!(newnode->datum = malloc(1 + strlen(node->datum)))) {
         free(newnode);
         newnode = NULL;
      }
      else {
         strcpy(newnode->datum, node->datum);
         newnode->next = NULL;
         newnode->linenum = 0;
         newnode->duplicates = 0;
      }
   }
   return newnode;
} /* hdupex */

/* ------------------- */

/* Unfortunately if the malloc package free operation is bad
   enough we are going to run into the O(N*N) behaviour when
   freeing all the hash table storage.  So be it. */
static void hfreewd(void *item)
{
   struct slist *node = item;

   free(node->datum);
   free(node);
   return;
} /* hfreex */

/* ------------------- */

static double evalwordhashing(hshtbl *h, struct slist *root)
{
   clock_t       cstart, cend;
   struct slist *node, currentwd, *inserted;
   char          buffer[MAXWORD];
   char         *ptr;
   int           ix, dupwords;

   currentwd.datum = &buffer[0];
   dupwords = 0;
   cstart = clock();
   node = root;
   while (node) {
      ptr = node->datum;
      while (1 == sscanf(ptr, "%s%n", buffer, &ix)) {
         ptr += ix;
         /* Note that for this table we do a deep copy */
         inserted = hshinsert(h, &currentwd);
         /* The following is peculiar to this test program */
         if (inserted->duplicates) dupwords++;
         inserted->duplicates++;
      }
      node = node->next;
   }
   cend = clock();
   printf("  %d duplicated words\n", dupwords);
   showstats(h);
   return (double)(cend - cstart) / CLOCKS_PER_SEC;
} /* evalwordhashing */

/* ------------------- */

int main(int argc, char **argv)
{
   hshtbl *thetable;

   if (akeyboard(stdin)) help();
   else {
      filedata = NULL;
      puts("\nTime for some elementary operations:");
      printf("Input %lu lines in %.3f secs\n",
              lines, timestrinput(stdin));
      printf("A null hash function call %s %.3f secs\n",
             "nullfn", timehashing(filedata, nullfn));
      printf("Hashing with %s in %.3f secs\n",
             "sx31hsh", timehashing(filedata, sx31hsh));
      printf("Hashing with %s in %.3f secs\n",
             "ssfh_ph", timehashing(filedata, ssfh_ph));

      puts("\nStatistics for storing full lines:");

      thetable = hashopen(sx31hsh, ssfh_ph, hdupex, hfreex);
      printf("Inserted with %s and %s in %.3f secs\n",
             "sx31hsh", "ssfh_ph",
             evaluatehashing(thetable, filedata));
      hshkill(thetable);

      thetable = hashopen(ssfh_ph, sx31hsh, hdupex, hfreex);
      printf("Inserted with %s and %s in %.3f secs\n",
             "ssfh_ph", "sx31hsh",
             evaluatehashing(thetable, filedata));
      hshkill(thetable);

      thetable = hashopen(sx37hsh, sx31hsh, hdupex, hfreex);
      printf("Inserted with %s and %s in %.3f secs\n",
             "sx37hsh", "sx31hsh",
             evaluatehashing(thetable, filedata));
      hshkill(thetable);

      puts("\nTime to parse words from the source:");
      printf("Parsing words in %.3f secs\n",
              timeparsewords(filedata));
      printf("Parsing and hashing words with %s in %.3f secs\n",
             "sx31hsh", timeparsehashwds(filedata, sx31hsh));
      printf("Parsing and hashing words with %s in %.3f secs\n",
             "ssfh_h", timeparsehashwds(filedata, ssfh_ph));
      printf("Parsing and hashing words with %s in %.3f secs\n",
             "sx37hsh", timeparsehashwds(filedata, sx37hsh));

      puts("\nStatistics for "
           "parsing, hashing and storing those words:");

      thetable = hashopen(sx31hsh, ssfh_ph, hdupewd, hfreewd);
      printf("Inserted with %s and %s in %.3f secs\n",
             "sx31hsh", "ssfh_ph",
             evalwordhashing(thetable, filedata));
      hshkill(thetable);

      thetable = hashopen(ssfh_ph, sx31hsh, hdupewd, hfreewd);
      printf("Inserted with %s and %s in %.3f secs\n",
             "ssfh_ph", "sx31hsh",
             evalwordhashing(thetable, filedata));
      hshkill(thetable);

      thetable = hashopen(sx37hsh, sx31hsh, hdupewd, hfreewd);
      printf("Inserted with %s and %s in %.3f secs\n",
             "sx37hsh", "sx31hsh",
             evalwordhashing(thetable, filedata));
      hshkill(thetable);
   }
   return 0;
} /* hshftst main */
